import inspect
import logging
import os
import re
import time
from base64 import b64encode
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from threading import Timer
from typing import Union

from binaryornot.check import is_binary
from box import Box, BoxKeyError

from .validator import Validator

FILE_MODES = ['binary', 'text', 'auto']

HTTP_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE', 'PATCH']


logger = logging.getLogger(__name__)


class EvaluationError(Exception):
    pass


class StopCycleError(Exception):
    pass


def make_list(item_or_items):
    """
    Makes a list out of the given items.
    Examples:
        >>> make_list(1)
        [1]
        >>> make_list('str')
        ['str']
        >>> make_list(('i', 'am', 'a', 'tuple'))
        ['i', 'am', 'a', 'tuple']
        >>> print(make_list(None))
        None
        >>> # An instance of lists is unchanged
        >>> l = ['i', 'am', 'a', 'list']
        >>> l_res = make_list(l)
        >>> l_res
        ['i', 'am', 'a', 'list']
        >>> l_res is l
        True

    Args:
        item_or_items: A single value or an iterable.
    Returns:
        Returns the given argument as an list.
    """
    if item_or_items is None:
        return None
    if isinstance(item_or_items, list):
        return item_or_items
    if isinstance(item_or_items, dict):
        return [item_or_items]
    if hasattr(item_or_items, '__iter__') and not isinstance(item_or_items, str):
        return list(item_or_items)
    return [item_or_items]


def camel_to_snake(name):
    """
    Converts camelCase to snake_case.
    https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case

    >>> camel_to_snake('CamelCase')
    'camel_case'
    >>> camel_to_snake('CamelCamelCase')
    'camel_camel_case'
    >>> camel_to_snake('Camel2Camel2Case')
    'camel2_camel2_case'
    >>> camel_to_snake('getHTTPResponseCode')
    'get_http_response_code'
    >>> camel_to_snake('get2HTTPResponseCode')
    'get2_http_response_code'
    >>> camel_to_snake('HTTPResponseCode')
    'http_response_code'
    >>> camel_to_snake('HTTPResponseCodeXYZ')
    'http_response_code_xyz'

    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def wildcards_to_regex(wildcard_patterns):
    """
    Examples:

        >>> import fnmatch, re
        >>> (wildcards_to_regex(['sensor.*', 'sensor.lamp'])
        ...     == [re.compile(fnmatch.translate('sensor.*')), re.compile(fnmatch.translate('sensor.lamp'))])
        True
        >>> wildcards_to_regex('sensor.lamp') == re.compile(fnmatch.translate('sensor.lamp'))
        True
        >>> print(wildcards_to_regex(None))
        None
        >>> wildcards_to_regex([])
        []
    """

    if not wildcard_patterns:
        return wildcard_patterns
    import fnmatch
    import re
    if is_iterable_but_no_str(wildcard_patterns):
        return [re.compile(fnmatch.translate(item)) for item in wildcard_patterns]
    Validator.is_instance(str, wildcard_patterns=wildcard_patterns)
    return re.compile(fnmatch.translate(wildcard_patterns))


def include_or_exclude(item, include_regex=None, exclude_regex=None):
    """

    Returns:
        True if the item should be included; False if the item should be excluded.

    Examples:
        >>> re_all = wildcards_to_regex('sensor.*')
        >>> re1 = wildcards_to_regex('sensor.light1')
        >>> re2 = wildcards_to_regex('sensor.light2')

        >>> include_or_exclude('sensor.light1', [re1, re2])
        True
        >>> include_or_exclude('sensor.light1', [re1, re2], [re_all])
        False
        >>> include_or_exclude('sensor.light1', None, [re1])
        False
        >>> include_or_exclude('sensor.light2', None, [re1])
        True
        >>> include_or_exclude('sensor.light2', [re1], None)
        False

    """
    for regex in exclude_regex or []:
        if regex.match(item):
            return False
    for regex in include_regex or []:
        if regex.match(item):
            return True

    return include_regex is None


def transform_dict_items(dct, keys_fun=None, vals_fun=None):
    """
    Transforms keys and/or values of the given dictionary by applying the given function.

    >>> dct = {"CamelCase": "gnaaa", "foo_oool": 42}
    >>> transform_dict_items(dct, keys_fun=camel_to_snake) == {"camel_case": "gnaaa", "foo_oool": 42}
    True
    >>> transform_dict_items(dct, vals_fun=str) == {"CamelCase": "gnaaa", "foo_oool": "42"}
    True
    >>> transform_dict_items(dct, keys_fun=camel_to_snake, vals_fun=str) == {"camel_case": "gnaaa", "foo_oool": "42"}
    True
    """
    if not keys_fun and not vals_fun:
        return dct

    def apply(kv, fun):
        return kv if not fun else fun(kv)

    return {apply(k, keys_fun): apply(v, vals_fun) for k, v in dct.items()}


def is_iterable_but_no_str(candidate):
    """
    Checks if the given candidate is an iterable but not a str instance

    Example:
        >>> is_iterable_but_no_str(['a'])
        True
        >>> is_iterable_but_no_str('a')
        False
        >>> is_iterable_but_no_str(None)
        False
    """
    return hasattr(candidate, '__iter__') and not isinstance(candidate, (str, bytes))


def interruptible_sleep(wait, callback, interval=0.5):
    """

    Waits the specified amount of time. The waiting can be interrupted when the callback raises a `StopCycleError`.
    The argument `interval` defines after how much wait time the callback should be called.

    Examples:

        >>> interruptible_sleep(0.5, lambda: print("Cycle"), 0.2)  # Does perform two cycles of 0.2 secs and one 0.1
        Cycle
        Cycle
        >>> i = 0
        >>> def callback():
        ...     global i
        ...     if i == 2:
        ...         raise StopCycleError()
        ...     print("Cycle")
        ...     i = i + 1
        >>> interruptible_sleep(0.5, callback, 0.1)  # Aborts the wait after 2 cycles
        Cycle
        Cycle

    Args:
        wait (float): Wait for this amount of time, if not interrupted (some semantias as in `time.sleep(wait)`.
        interval (float): How often the callback should be called.

    Returns:
        None
    """
    wait = float(wait)
    interval = float(interval)

    complete_cycles = int(wait // interval)
    try:
        for i in range(0, complete_cycles):
            callback()  # Should raise a StopCycleError error when waiting should be aborted
            time.sleep(interval)

        time.sleep(wait % interval)
    except StopCycleError:
        pass


def make_public_protected_private_attr_lookup(attr_name, as_dict=False):
    """
    Given an attribute name this function will generate names of public, private and protected attribute names.
    The order is of lookups is always the given attr_name first and then descending by visibility (public -> proctected
    -> private)

    Examples:
        >>> make_public_protected_private_attr_lookup('my_lookup')  # Public attribute name
        ['my_lookup', '_my_lookup', '__my_lookup']
        >>> make_public_protected_private_attr_lookup('_my_lookup')  # Protected attribute name
        ['_my_lookup', 'my_lookup', '__my_lookup']
        >>> make_public_protected_private_attr_lookup('__my_lookup')  # Private attribute name
        ['__my_lookup', 'my_lookup', '_my_lookup']
        >>> l = make_public_protected_private_attr_lookup('_my_lookup', as_dict=True)
        >>> list(l.keys()), list(l.values())
        (['protected', 'public', 'private'], ['_my_lookup', 'my_lookup', '__my_lookup'])

    """
    Validator.is_instance(str, lookup_name=attr_name)
    as_dict = try_parse_bool(as_dict, default=False)
    if attr_name.startswith('__'):
        # __lookup, lookup, _lookup
        res = OrderedDict([('private', attr_name), ('public', attr_name[2:]), ('protected', attr_name[1:])])
    elif attr_name.startswith('_'):
        # _lookup, lookup, __lookup
        res = OrderedDict([('protected', attr_name), ('public', attr_name[1:]), ('private', '_' + attr_name)])
    else:
        # lookup, _lookup, __lookup
        res = OrderedDict([('public', attr_name), ('protected', '_' + attr_name), ('private', '__' + attr_name)])
    return res if as_dict else list(res.values())


def instance_lookup(instance, lookups):
    """
    Will lookup the instance for the given attribute names in `lookups`.
    Returns the first occurrence found.

    Examples:

        >>> class Foo:
        ...     def __init__(self):
        ...         self.a = 1
        ...         self.b = 2
        ...         self.c = 3
        ...         self.__a = 99
        >>> instance = Foo()
        >>> instance_lookup(instance, 'a')
        1
        >>> instance_lookup(instance, '__a')
        99
        >>> instance_lookup(instance, ['c', 'b'])
        3
        >>> instance_lookup(instance, ['d', 'b'])
        2
        >>> print(instance_lookup(instance, ['x', 'y', 'z']))
        None

    Args:
        instance: The class to lookup for the given attribute names.
        lookups: Attribute names to lookup.

    Returns:
        Returns the value of the first found attribute name in lookups. If none is found, simply None is returned.s
    """
    lookups = make_list(lookups)
    # There might be some name mangling for private attributes - add them as lookups
    privates = ['_{classname}{attrname}'.format(classname=type(instance).__name__, attrname=x)
                for x in lookups if x.startswith('__') and not x.endswith('__')]
    lookups = lookups + privates
    _Nothing = object()
    for lookup in lookups:
        val = getattr(instance, lookup, _Nothing)
        if val is not _Nothing:
            return val
    return None


def safe_eval(source, **context):
    """
    Calls the `eval` function with modified globals() and locals(). By default of `eval` the current globals() and
    locals() are passed. We do not want that behaviour because the user can make a lot of nasty things with it
    (think of a already imported `os` module...).

    Instead we pass some functions (not all) from `__builtins__` (aka the whitelist) and remove all other modules.
    The caller can pass some additional context (like variables) to the eval.

    Examples:

        >>> payload = {'foo': {'bar': 1, 'baz': 2}}
        >>> safe_eval("str(payload['foo']['bar'])", payload=payload, str=str)  # We pass some ctx the eval has access to
        '1'
        >>> import os
        >>> safe_eval("os.system('echo 0')")  # Try to access os
        Traceback (most recent call last):
        ...
        pnp.utils.EvaluationError: Failed to evaluate 'os.system('echo 0')'
        >>> safe_eval("os.system('echo 0')", os=os)  # The caller can pass the os module... if he really wants to...
        0

    Args:
        source: The source to execute.
        **context: Additional context (e.g. variables) to pass to eval.

    Returns:
        Return the result of the eval. If something went wrong it raises an error.
    """
    try:
        return eval(str(source), {'__builtins__': {}}, context)
    except Exception as exc:
        raise EvaluationError("Failed to evaluate '{source}'".format(**locals())) from exc


def try_parse_int(candidate):
    """
    Convert the given candidate to int. If it fails None is returned.

    Example:

        >>> type(try_parse_int(1))  # int will still be int
        <class 'int'>
        >>> type(try_parse_int("15"))
        <class 'int'>
        >>> print(try_parse_int("a"))
        None

    Args:
        candidate: The candidate to convert.

    Returns:
        Returns the converted candidate if convertible; otherwise None.

    """
    try:
        return int(candidate)
    except (ValueError, TypeError):
        return None


def try_parse_int_float_str(candidate):
    """
    Tries to parse the given value as an int or float. If this does not work simply the string-representation of the
    candidate will be returned.

    Examples:
        >>> res = try_parse_int_float_str(5)
        >>> res, type(res)
        (5, <class 'int'>)
        >>> res = try_parse_int_float_str(5.5)
        >>> res, type(res)
        (5.5, <class 'float'>)
        >>> res = try_parse_int_float_str(5.0)
        >>> res, type(res)
        (5, <class 'int'>)
        >>> res = try_parse_int_float_str('simple_string')
        >>> res, type(res)
        ('simple_string', <class 'str'>)

    Args:
        candidate: Candidate to convert.

    """
    try:
        ffloat = float(candidate)
        return int(ffloat) if ffloat.is_integer() else ffloat
    except:
        return str(candidate)


def try_parse_bool(value, default=None):
    """
    Tries to parse the given value as a boolean. If the parsing is unsuccessful the default will be returned.

    Args:
        value: Value to parse.
        default (optional): The value to return in case the conversion is not successful.

    Returns:
        If successful the converted representation of value; otherwise the default.

    Examples:
        >>> print(try_parse_bool(1))
        True
        >>> print(try_parse_bool('true'))
        True
        >>> print(try_parse_bool('T'))
        True
        >>> print(try_parse_bool('unknown', default='Unknown'))
        Unknown
        >>> print(try_parse_bool(None, default=True))
        True
        >>> print(try_parse_bool(1.0))
        True
        >>> print(try_parse_bool(0.99))
        True
        >>> print(try_parse_bool(0.0))
        False
        >>> print(try_parse_bool(lambda x: True, default="default"))
        True
    """
    if value is None:
        return default
    if isinstance(value, bool):
        return value
    if isinstance(value, str):
        if value.lower() in ['t', 'true', '1', 'wahr', 'ja', 'yes', 'on']:
            return True
        elif value.lower() in ['f', 'false', '0', 'falsch', 'nein', 'no', 'off']:
            return False
        else:
            return default

    try:
        return bool(value)
    except (ValueError, TypeError):
        return default


def on_off(value):
    """
    Tries to convert the given value to a on / off literal. If conversion is not possible (unsupported datatype) it will
    silently fail and return 'off'.

    Examples:

        >>> on_off(True), on_off(False)
        ('on', 'off')
        >>> on_off('True'), on_off('False')
        ('on', 'off')
        >>> on_off(None)
        'off'
        >>> on_off(99), on_off(0)
        ('on', 'off')

    Args:
        value: The value to convert to on / off.

    Returns:
        Return 'on' if the the `value` is True or equivalent (see try_parse_bool); otherwise 'off'.
    """
    return 'on' if try_parse_bool(value, default=False) else 'off'


def get_file_mode(file_path, mode):
    """
    Returns 'rb' if mode = 'binary'.
    Returns 'r' if mode = 'text'
    Returns 'rb' or 'r' if mode = 'auto' -> Will be automatically detected.

    Example:

        >>> get_file_mode("doesn't matter", 'binary')
        'rb'
        >>> get_file_mode("doesn't matter", 'text')
        'r'
        >>> get_file_mode(__file__, 'auto')
        'r'

    Args:
        file_path: File to load.
        mode: One of ['binary', 'text', 'auto'].

    Returns:
        One of ['rb', 'r']

    """
    if mode == 'binary':
        return 'rb'
    if mode == 'text':
        return 'r'
    if mode == 'auto':
        return 'rb' if is_binary(file_path) else 'r'

    raise ValueError("Argument 'mode' is expected to be one of: auto, binary, text")


def load_file(fp, mode='auto', base64=False):
    if not isinstance(base64, bool):
        raise TypeError("Argument 'base64' is expected to be a bool, but is {}".format(type(base64)))
    mode = 'binary' if base64 else mode
    if mode not in FILE_MODES:
        raise ValueError("Argument 'mode' is expected to be one of: {}".format(FILE_MODES))

    read_mode = get_file_mode(fp, mode=mode)
    with open(fp, read_mode) as fs:
        contents = b64encode(fs.read()) if base64 else fs.read()

    return dict(file_name=os.path.basename(fp), content=contents, mode=read_mode, base64=base64)


def get_first_existing_file(*file_list):
    """
    Given is a list of possible files. The function returns the first item that exists. If none of the candidates
    exists `None` is returned.

    Examples:
        >>> import tempfile
        >>> with tempfile.NamedTemporaryFile() as ntf:
        ...     res = get_first_existing_file('idonotexists.exists', ntf.name, '/asd/dasd/ad.ad')
        >>> res == ntf.name
        True
    """
    for fp in file_list:
        if isinstance(fp, str) and os.path.isfile(fp):
            return fp
    return None


def get_bytes(stream_or_file):
    """
    Returns the bytes of the given stream or file argument. If it is a file, it will be opened in binary mode.
    Examples:
        >>> import tempfile
        >>> tmp = tempfile.NamedTemporaryFile()
        >>> print(get_bytes(tmp.name))
        b''
        >>> with open(tmp.name, 'rb') as stream:
        ...     get_bytes(stream)
        b''
        >>> get_bytes('thatonedoesnotexist')
        Traceback (most recent call last):
            ...
        ValueError: Argument 'stream_or_file' is neither a stream nor a file

    Args:
        stream_or_file: Stream or file to read the bytes from.

    Returns:
        Bytes of ``stream_or_file``
    """
    if getattr(stream_or_file, 'read', None):
        return stream_or_file.read()
    from pathlib import Path
    if Path(stream_or_file).is_file():
        with open(stream_or_file, 'rb') as file:
            return file.read()

    raise ValueError("Argument 'stream_or_file' is neither a stream nor a file")


DurationLiteral = Union[str, int]


def parse_duration_literal(literal: DurationLiteral):
    """
    Converts duration literals as '1m', '1h', and so on to an actual duration in seconds.
    Supported are 's' (seconds), 'm' (minutes), 'h' (hours), 'd' (days) and 'w' (weeks).

    Examples:

        >>> parse_duration_literal(60)  # Int will be interpreted as seconds
        60
        >>> parse_duration_literal('10')  # Any int convertible will be interpreted as seconds
        10
        >>> parse_duration_literal('20s')  # Seconds literal
        20
        >>> parse_duration_literal('2m')  # Minutes literal
        120
        >>> parse_duration_literal('1h')  # Hours literal
        3600
        >>> parse_duration_literal('1d')  # Days literal
        86400
        >>> parse_duration_literal('1w')  # Weeks literal
        604800
        >>> parse_duration_literal('invalid')  # Invalid will raise an error
        Traceback (most recent call last):
        ...
        TypeError: Interval 'invalid' is not a valid literal

    Args:
        literal: Literal to parse.

    Returns:
        Returns the converted literal's duration in seconds. If conversion is not possible
        an exception is raised.
    """
    try:
        # if successful we got seconds
        return int(literal)
    except:  # pylint: disable=broad-except
        # We have to check for s, m, h, d, w suffix
        seconds_per_unit = {"s": 1, "m": 60, "h": 3600, "d": 86400, "w": 604800}
        # Remove all non-alphanumeric letters
        s = re.sub('[^0-9a-zA-Z]+', '', str(literal))
        value_str, unit = s[:-1], s[-1].lower()
        value = try_parse_int(value_str)
        if value is None or unit not in seconds_per_unit:
            raise TypeError("Interval '{}' is not a valid literal".format(literal))
        return value * seconds_per_unit[unit]


def safe_get(dct, *keys, default=None):
    """
    Get the value inside the dictionary that is accessible by the given keys or - if a key doesn't exists - the default.

    Examples:

        >>> dct = dict(nested=dict(nested1="a", nested2="b"), val="z")
        >>> safe_get(dct, "nested", "nested1")
        'a'
        >>> safe_get(dct, "nested", "unknown") is None
        True
        >>> safe_get(dct, "nested", "unknown", default="default")
        'default'
        >>> safe_get(dct) == dict(nested=dict(nested1="a", nested2="b"), val="z")
        True

    Args:
        dct: Dictionary to get the value from.
        *keys (list of hashable type): Keys to lookup.
        default (object, optional):

    Returns:
        Returns the requested value inside the probably nested dictionary. If a key does not exists, the function will
        return the default.
    """
    for key in keys:
        try:
            dct = dct[key]
        except KeyError:
            return default
    return dct


def get_field_mro(cls, field_name):
    """
    Goes up the mro (method resolution order) of the given class and returns the union of a given field.

    Example:

        >>> class Root:
        ...     __myfield__ = 'root'

        >>> class A(Root):
        ...     __myfield__ = ['a', 'common']

        >>> class B(Root):
        ...     __myfield__ = ['b', 'common']

        >>> class Final(A, B):
        ...     __myfield__ = 'final'

        >>> get_field_mro(Final, '__myfield__') == {'root', 'a', 'b', 'common', 'final'}
        True
        >>> get_field_mro(A, '__myfield__') == {'root', 'a', 'common'}
        True
        >>> f = Final()
        >>> get_field_mro(f, '__myfield__') == {'root', 'a', 'b', 'common', 'final'}
        True

    Args:
        cls:
        field_name:

    Returns:

    """
    res = set()
    if not hasattr(cls, '__mro__'):
        # cls might be an instance. mro is only available on classes
        if not hasattr(type(cls), '__mro__'):
            # No class, no instance ... Return empty set
            return res
        cls = type(cls)

    for c in inspect.getmro(cls):
        values_ = getattr(c, field_name, None)
        if values_ is not None:
            res = res.union(set(make_list(values_)))
    return res


def auto_str(__repr__=False):
    """
    Use this decorator to auto implement __str__() and optionally __repr__() methods on classes.

    Args:
        __repr__ (bool): If set to true, the decorator will auto-implement the __repr__() method as well.

    Returns:
        callable: Decorating function.

    Note:
        There are known issues with self referencing (self.s = self). Recursion will be identified by the python
        interpreter and will do no harm, but it will actually not work.
        A eval(class.__repr__()) will obviously not work, when there are attributes that are not part of the
        __init__'s arguments.

    Example:
        >>> @auto_str(__repr__=True)
        ... class Demo(object):
        ...    def __init__(self, i=0, s="a", l=None, t=None):
        ...        self.i = i
        ...        self.s = s
        ...        self.l = l
        ...        self.t = t
        >>> dut = Demo(10, 'abc', [1, 2, 3], (1,2,3))
        >>> print(dut.__str__())
        Demo(i=10, l=[1, 2, 3], s='abc', t=(1, 2, 3))
        >>> print(eval(dut.__repr__()).__str__())
        Demo(i=10, l=[1, 2, 3], s='abc', t=(1, 2, 3))
        >>> print(dut.__repr__())
        Demo(i=10, l=[1, 2, 3], s='abc', t=(1, 2, 3))
    """
    def decorator(cls):
        def __str__(self):
            items = ["{name}={value}".format(
                name=name,
                value=vars(self)[name].__repr__()
            ) for name in [key for key in sorted(vars(self))]
                if name not in get_field_mro(self.__class__, '__auto_str_ignore__')]
            return "{clazz}({items})".format(
                clazz=str(type(self).__name__),
                items=', '.join(items)
            )
        cls.__str__ = __str__
        if __repr__:
            cls.__repr__ = __str__

        return cls

    return decorator


def auto_str_ignore(ignore_list):
    """
    Use this decorator to suppress any fields that should not be part of the dynamically created
    `__str__` or `__repr__` function of `auto_str`.

    Example:

        >>> @auto_str()
        ... @auto_str_ignore(["l", "d"])
        ... class Demo(object):
        ...    def __init__(self, i=0, s="a", l=None, d=None):
        ...        self.i = i
        ...        self.s = s
        ...        self.l = l
        ...        self.d = d
        >>> dut = Demo(10, 'abc', [1, 2, 3], {'a': 1, 'b': 2})
        >>> print(str(dut))
        Demo(i=10, s='abc')

    Args:
        ignore_list: List or item of the fields to suppress by `auto_str`.

    Returns:
        Returns a decorator.
    """
    def decorator(cls):
        ignored = make_list(ignore_list)
        cls.__auto_str_ignore__ = ignored
        return cls
    return decorator


class classproperty(property):
    """
    Decorator classproperty:
    Make class methods look like read-only class properties.
    Writing to that classproperty will not do what you expect ;-)
    Examples:
        >>> class Foo(object):
        ...     _instance = 5
        ...     @classproperty
        ...     def my_prop(cls):
        ...         return cls._instance
        >>> Foo.my_prop
        5
        >>> Foo._instance
        5
        >>> Foo._instance = 15
        >>> Foo.my_prop
        15
        >>> Foo.my_prop = 10
        >>> Foo._instance
        15
    """
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class Loggable(object):
    """
    Adds a logger property to the class to provide easy access to a configured logging instance to use.
    Example:
        >>> class NeedsLogger(Loggable):
        ...     def do(self, message):
        ...         self.logger.info(message)
        >>> dut = NeedsLogger()
        >>> dut.do('mymessage')
    """
    @classproperty
    def logger(cls):
        """
        Configures and returns a logger instance for further use.
        Returns:
            (logging.Logger)
        """
        component = "{}.{}".format(cls.__module__, cls.__name__)
        return logging.getLogger(component)


class FallbackBox(Box):
    """
    In some cases we want to handle integers and floats as string when accessing
    a boxed dictionary. For example we would like to mimic access paths like the zway api can:
        devices[19].instances[0].commandClasses[49].data[3].val.value
    In this case all int looking ones are actual string keys. But for plugins we want it
    as easy as possible. So we first take the key as it is to access the data, but when it raises a `KeyError`
    we cast the key to str and try it again to access the key as a fallback.

    Examples:

        >>> d = {'devices': {'1': {'instances': {'0': 'instance0', 1: 'instance1'}}}}
        >>> dut = FallbackBox(d)
        >>> dut.devices[1].instances[0]  # devices.1 is a syntax error, cause 1 is not a valid identifier
        'instance0'
        >>> dut.devices[1].instances[1]
        'instance1'
    """
    def __getitem__(self, item, _ignore_default=False):
        try:
            return super().__getitem__(item, _ignore_default=_ignore_default)
        except BoxKeyError as ke:
            if isinstance(item, (int, float)):
                return super().__getitem__(str(item), _ignore_default=_ignore_default)
            raise ke


class Debounce:
    """
    Defers a function execution until `wait` seconds have elapsed since the last time it was invoked.
    It is a way to collect multiple invokes and only really execute the last invocation.

    Examples:

        >>> results = []
        >>> def foo(counter):
        ...     results.append(counter)

        >>> dut = Debounce(foo, 0.2)  # Defers the execution for 200 milliseconds
        >>> for cnt in range(3): dut(cnt)
        >>> time.sleep(0.3)
        >>> results
        [2]

        >>> for cnt in range(4): dut(cnt)
        >>> time.sleep(0.3)
        >>> results
        [2, 3]

        >>> results.clear()
        >>> for cnt in range(3): dut(cnt)
        >>> dut.execute_now()
        >>> for cnt in range(4): dut(cnt)
        >>> time.sleep(0.3)
        >>> results
        [2, 3]



    """
    def __init__(self, fun, wait=0.5):
        Validator.is_function(fun=fun)
        self.fun = fun
        self.wait = float(wait)
        self.timer = None

    def _safe_stop_timer(self):
        if self.timer is not None:
            self.timer.cancel()

    def execute_now(self):
        # Without checking self.timer.finished.is_set(), in rare situations the actual function would get executed
        # twice
        if self.timer is not None and not self.timer.finished.is_set():
            self._safe_stop_timer()
            self.timer.function(*self.timer.args, **self.timer.kwargs)

    def __call__(self, *args, **kwargs):
        self._safe_stop_timer()
        self.timer = Timer(self.wait, self.fun, args, kwargs)
        self.timer.start()


class Singleton(object):
    """
    Derive from this class to create a singleton.
    Examples:
        >>> class Foo(Singleton):
        ...     def __init__(self, state):
        ...         self.state = state
        >>> # Instance will be created on first call ...
        >>> f = Foo('statef')
        >>> # ... and passing different arguments will not change the already created instance.
        >>> g = Foo('stateg')
        >>> f is g
        True
        >>> print(f.state)
        statef
        >>> print(g.state)
        statef
        >>> # Two ways to retrieve the singleton instance
        >>> # 1. Pretend to "create" a new instance. This will return the already created instance.
        >>> Foo() is f
        True
        >>> # Arguments work as well (but will be ignored)
        >>> Foo("i", "get", "ignored") is f
        True
        >>> # 2. The more explicit variant: Call ``instance`` on the Singleton (Explicit is good).
        >>> Foo.instance is f
        True
        >>> # You can re-initialize the instance
        >>> f.init('new_state')
        >>> print(f.state)
        new_state
        >>> print(g.state)
        new_state
    """
    def __new__(cls, *args, **kwargs):
        instance = cls.__dict__.get("__singleton__")
        if instance is None:
            # Create the singleton
            cls.__singleton__ = instance = object.__new__(cls)
            # Initialize it once.
            instance.__init__(*args, **kwargs)

            # Backup initializer ...
            cls.init = cls.__init__
            # .. and then forget about it...

            def empty_init(self, *args, **kwargs):
                pass
            cls.__init__ = empty_init
        return instance

    @classproperty
    def instance(cls):
        """
        Returns the singleton instance. If there is none yet, it will be created by calling the ``init`` method
        without any arguments.
        """
        return cls.__new__(cls)


class FileLock(object):
    """
    A file locking mechanism that has context-manager support so you can use it within a with statement.
    This implementation should be relatively cross-platform compatible cause it doesn't rely on msvcrt or fcntl
    for the locking.
    You can nest calls to ``acquire`` it will only lock the file once and release it with the last call to ``release``.
    Examples:
        >>> import tempfile
        >>> tmp = tempfile.NamedTemporaryFile()
        >>> # Basic usage with context manager
        >>> with FileLock(tmp.name) as l:
        ...     # Do something nice with the file
        ...     print(l.locked)
        True
        >>> # Basic usage w/o context manager
        >>> l = FileLock(tmp.name)
        >>> l.locked
        False
        >>> l.acquire()
        >>> try:
        ...     # Do something nice with the file
        ...     print(l.locked)
        ... finally:
        ...     l.release()
        True
        >>> print(l.locked)
        False
        >>> # Lock counter
        >>> l = FileLock(tmp.name)
        >>> with l:
        ...     print(l.lock_counter)
        ...     with l:
        ...         print(l.lock_counter)
        ...     print((l.lock_counter, l.locked))
        1
        2
        (1, True)
        >>> print((l.lock_counter, l.locked))
        (0, False)
        >>> # Blocking
        >>> with FileLock(tmp.name) as l1:
        ...     with FileLock(tmp.name, timeout=1) as l2:
        ...         pass
        Traceback (most recent call last):
        ...
        pnp.utils.FileLock.TimeoutError: Timeout occurred
    """

    class TimeoutError(Exception):
        pass

    def __init__(self, file_name, timeout=10, delay=.05):
        """
        Prepare the file locker. Specify the file to lock and optionally the maximum timeout (seconds) and the delay
        between each attempt to lock.
        Args:
            file_name: The file to lock exclusively.
            timeout: If the file is already locked by another process, this instance tries to acquire the lock
                for a maximum of `timeout` seconds.
            delay:
        """
        self.lock_counter = 0
        self.lockfile = os.path.join(os.getcwd(), "%s.lock" % file_name)
        self.file_name = file_name
        self.timeout = timeout
        self.delay = delay
        self.fd = None

    def acquire(self):
        """
        Acquires the lock, if possible. If the file is already locked, the algorithm tries to acquire the lock until
        it either gets the lock or the timeout threshold is exceeded. In case of a timeout an error is raised.
        Returns:
            None
        """
        if self.lock_counter > 0:
            # Dude, you already locked the file...
            self.lock_counter += 1
            return

        start_time = time.time()
        while True:
            try:
                self.fd = os.open(self.lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                self.lock_counter += 1
                break
            except OSError as e:
                import errno
                if e.errno != errno.EEXIST:
                    raise
                if (time.time() - start_time) >= self.timeout:
                    raise FileLock.TimeoutError("Timeout occurred")
                time.sleep(self.delay)

    def release(self):
        """
        Unlocks the file when the exclusive lock is no longer needed.
        Returns:
            None
        """
        if self.lock_counter <= 0:
            # Dude, there is no lock...
            return

        self.lock_counter -= 1
        if self.lock_counter == 0:
            os.close(self.fd)
            os.unlink(self.lockfile)

    @property
    def locked(self):
        """
        Checks whether the file is currently locked by any process.
        Returns: True if the file is currently locked (by any process, not just this one); otherwise False.
        """
        return os.path.isfile(self.lockfile)

    def __enter__(self):
        """
        Context manager acquire lock.
        Returns:
            Self
        """
        self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        """
        Context manager release.
        Args:
            type: unused
            value: unused
            traceback: unused
        Returns:
            None
        """
        self.release()

    def __del__(self):
        """
        Make sure that this instance doesn't leave a lockfile lying around when garbage collected.
        Returns:
            None
        """
        self.release()


class Parallel:
    """
    Provides an interface to easily execute functions in parallel.

    Examples:

        >>> def long_fun(r):
        ...     return r
        >>> p = Parallel(workers=2)
        >>> p(long_fun, 1, fun_key=1)
        >>> p(long_fun, 2, fun_key=2)
        >>> p(long_fun, 3, fun_key=3)
        >>> p.run_until_complete()
        ['finished', 'finished', 'finished']
        >>> p.results == {1: 1, 2: 2, 3: 3}
        True
        >>> p(long_fun, 4, fun_key=4)
        >>> p.run_until_complete()
        ['finished', 'finished', 'finished', 'finished']
        >>> p.results == {1: 1, 2: 2, 3: 3, 4: 4}
        True

        >>> memory = set()
        >>> p = Parallel(workers=2)
        >>> p(long_fun, 1, callback=lambda res: memory.add(res()))
        >>> p(long_fun, 2, callback=lambda res: memory.add(res()))
        >>> p(long_fun, 3, callback=lambda res: memory.add(res()))
        >>> p.run_until_complete()
        ['finished', 'finished', 'finished']
        >>> memory == {1, 2, 3}
        True
    """
    def __init__(self, workers=2):
        self.funcs = []
        self.workers = int(workers)
        if self.workers <= 0:
            self.workers = 1
        self._results = {}

    @property
    def results(self):
        import copy
        return copy.copy(self._results)

    def __call__(self, fun, *args, callback=None, fun_key=None, **kwargs):
        Validator.is_function(fun=fun)
        Validator.is_function(allow_none=True, callback=callback)
        self.funcs.append((fun, callback, fun_key, args, kwargs))

    def _intercept(self, future, callback, fun_key):
        if fun_key:
            try:
                self._results[fun_key] = future.result()
            except:
                import traceback
                logger.warning("Parallel result raised an error\n{}".format(traceback.format_exc()))
        if callback:
            try:
                callback(future.result)
            except:
                import traceback
                logger.warning("Parallel callback raised an error\n{}".format(traceback.format_exc()))

    def run_until_complete(self):
        self._results.clear()
        futures = []
        with ThreadPoolExecutor(max_workers=self.workers) as executor:
            for fun, cb, fun_key, args, kwargs in self.funcs:
                future = executor.submit(fun, *args, **kwargs)
                future.add_done_callback(partial(self._intercept, callback=cb, fun_key=fun_key))
                futures.append(future)

        return ['finished' if f.exception() is None else 'error' for f in futures]



import multiprocessing
import threading

from . import Engine, RetryHandler, PushExecutor, NotSupportedError, SimpleRetryHandler, NoRetryHandler
from ..models import TaskSet
from ..validator import Validator


class SequentialEngine(Engine):
    """Sequential engine that only allows to run a single pull. It is more like a script execution and nothing
    is concurrent."""
    def __init__(self, retry_handler=None):
        super().__init__()
        Validator.is_instance(RetryHandler, allow_none=True, retry_handler=retry_handler)
        self.retry_handler = retry_handler
        if self.retry_handler is None:
            self.retry_handler = SimpleRetryHandler()
        self.stopped = multiprocessing.Event()
        self._task = None

    def _run(self, tasks: TaskSet):
        if len(tasks) < 1:
            return
        if len(tasks) > 1:
            raise NotSupportedError("The Sequential engine can only handle one task")

        task = tasks[list(tasks.keys())[0]]
        self._task = task

        def on_payload(plugin, payload):
            for push in task.pushes:
                try:
                    PushExecutor().execute('sequential', payload, push)
                except:  # pylint: disable=broad-except
                    import traceback
                    self.logger.error("\n{}".format(traceback.format_exc()))

        task.pull.instance.on_payload = on_payload
        task.pull.instance._stopped = self.stopped

        while not self.stopped.is_set():
            try:
                task.pull.instance.pull()
                if isinstance(self.retry_handler, NoRetryHandler):
                    self.stopped.set()

                if not self.stopped.is_set():
                    # Bad thing... Pulling exited unexpectedly
                    self.logger.error("Pulling of '{pull}' exited unexpectedly".format(
                        thread=threading.get_ident(),
                        pull=task.pull.instance,
                    ))
            except KeyboardInterrupt:
                self.stopped.set()
            except:   # pylint: disable=broad-except
                # Bad thing... Pulling exited with exception
                import traceback
                self.logger.error("Pulling of '{pull}' raised an error\n{error}".format(
                    thread=threading.get_ident(),
                    pull=task.pull.instance,
                    error=traceback.format_exc()
                ))
            finally:
                if not self.stopped.is_set():
                    directive = self.retry_handler.handle_error()
                    if directive.abort:
                        self.logger.error("Pulling of '{pull}' exited due to retry limitation".format(
                            thread=threading.get_ident(),
                            pull=task.pull.instance,
                        ))
                        self.stopped.set()
                    else:
                        self._sleep(directive.wait_for)

    def _stop(self):
        self.stopped.set()



import os
import pathlib
from functools import partial
from ruamel import yaml

import schema

from . import Polling
from .. import load_optional_module
from ...utils import auto_str_ignore, camel_to_snake, transform_dict_items, make_list, FileLock, Parallel
from ...validator import Validator


@auto_str_ignore(['_tokens', '_client', '_tokens_tstamp', '_client_lock'])
class _FitbitBase(Polling):
    __prefix__ = 'fitbit'

    EXTRA = 'fitbit'

    TOKEN_SCHEMA = schema.Schema({
        'client_id': schema.Use(str),
        'client_secret': schema.Use(str),
        'access_token': schema.Use(str),
        'refresh_token': schema.Use(str),
        'expires_at': schema.Use(float)
    })

    def __init__(self, config, system=None, **kwargs):
        super().__init__(**kwargs)
        self._system = system

        # Absolute config path? Relative -> relative to pnp configuration file
        self._config = config
        if not os.path.isabs(config):
            self._config = os.path.join(self.base_path, config)
        Validator.is_file(config=self._config)

        import threading
        self._client_lock = threading.Lock()
        self._tokens = None  # Set by _load_tokens
        self._tokens_tstamp = None  # Set by _load_tokens and _save_tokens
        self._client = None  # Set by client property

    @property
    def client(self):
        with self._client_lock:
            if self._load_tokens():  # Check if tokens got refreshed by another process
                fb = load_optional_module('fitbit', self.EXTRA)
                self._client = fb.Fitbit(**self._tokens, oauth2=True, refresh_cb=self._save_tokens, system=self._system)
            return self._client

    def _load_tokens(self):
        current_tstamp = pathlib.Path(self._config).stat().st_mtime
        # Compare memorized modified timestamp (might be None initially) and current
        if self._tokens_tstamp == current_tstamp:
            # Tokens did not change. Skip
            return False

        self.logger.debug("[{self.name}] Loading tokens from {self._config}: Requesting lock".format(**locals()))
        with FileLock(self._config):
            self.logger.debug("[{self.name}] Loading tokens from {self._config}: Lock acquired".format(**locals()))
            with open(self._config, 'r') as fp:
                _tokens = yaml.safe_load(fp)
                self._tokens = self.TOKEN_SCHEMA.validate(_tokens)
            self._tokens_tstamp = current_tstamp
        self.logger.debug("[{self.name}] Loading tokens from {self._config}: Lock released".format(**locals()))
        return True

    def _save_tokens(self, tokens):
        new_config = {
            'client_id': self._tokens.get('client_id'),
            'client_secret': self._tokens.get('client_secret'),
            'access_token': tokens.get('access_token'),
            'refresh_token': tokens.get('refresh_token'),
            'expires_at': tokens.get('expires_at')
        }
        self.logger.debug("[{self.name}] Saving tokens to {self._config}: Requesting lock".format(**locals()))
        self.TOKEN_SCHEMA.validate(new_config)
        with FileLock(self._config):
            self.logger.debug("[{self.name}] Saving tokens to {self._config}: Lock acquired".format(**locals()))
            with open(self._config, 'w') as fp:
                yaml.dump(new_config, fp, default_flow_style=False)
            self._tokens_tstamp = pathlib.Path(self._config).stat().st_mtime
            self._tokens = new_config
            self.logger.debug("[{self.name}] Saving tokens to {self._config}: Lock released".format(**locals()))

    def poll(self):
        raise NotImplementedError()  # pragma: no cover


class Devices(_FitbitBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def poll(self):
        return [transform_dict_items(d, keys_fun=camel_to_snake) for d in self.client.get_devices()]


@auto_str_ignore(['_goals_map'])
class Goal(_FitbitBase):
    def __init__(self, goals, **kwargs):
        super().__init__(**kwargs)
        self._goals_map = self._create_goal_map()
        self._goals = make_list(goals)
        for goal in self._goals:
            Validator.one_of(list(self._goals_map.keys()), goal=goal)

    def _create_goal_map(self):
        return {
            'body/fat': lambda goal: float(self.client.body_fat_goal().get('goal', {}).get('fat')),
            'body/weight': lambda goal: float(self.client.body_weight_goal().get('goal', {}).get('weight')),
            'activities/daily/activeMinutes': self._get_daily_activity_goal,
            'activities/daily/caloriesOut': self._get_daily_activity_goal,
            'activities/daily/distance': self._get_daily_activity_goal,
            'activities/daily/floors': self._get_daily_activity_goal,
            'activities/daily/steps': self._get_daily_activity_goal,
            'activities/weekly/distance': self._get_weekly_activity_goal,
            'activities/weekly/floors': self._get_weekly_activity_goal,
            'activities/weekly/steps': self._get_weekly_activity_goal,
            'foods/calories': lambda goal: int(self.client.food_goal().get('goals', {}).get('calories')),
            'foods/water': lambda goal: int(self.client.water_goal().get('goal', {}).get('goal'))
        }

    def _get_daily_activity_goal(self, goal):
        single_goal = goal.split('/')[-1]
        conv = float if single_goal == 'distance' else int
        return conv(self.client.activities_daily_goal().get('goals', {}).get(single_goal))

    def _get_weekly_activity_goal(self, goal):
        single_goal = goal.split('/')[-1]
        conv = float if single_goal == 'distance' else int
        return conv(self.client.activities_weekly_goal().get('goals', {}).get(single_goal))

    def _call(self, goal):
        return self._goals_map.get(goal)(goal)

    def poll(self):
        p = Parallel(workers=4)
        for goal in self._goals:
            p(self._call, fun_key=goal, goal=goal)
        p.run_until_complete()
        return transform_dict_items(p.results, keys_fun=camel_to_snake)


@auto_str_ignore(['_resource_map'])
class Current(_FitbitBase):
    def __init__(self, resources, **kwargs):
        super().__init__(**kwargs)
        self._resource_map = self._create_resource_map()
        self._resources = make_list(resources)
        for res in self._resources:
            Validator.one_of(list(self._resource_map.keys()), resource=res)

    def _create_resource_map(self):
        return {
            # Activities
            'activities/calories': partial(self._conv, conv_fun=int),
            'activities/caloriesBMR': partial(self._conv, conv_fun=int),
            'activities/steps': partial(self._conv, conv_fun=int),
            'activities/distance': partial(self._conv, conv_fun=float),
            'activities/floors': partial(self._conv, conv_fun=int),
            'activities/elevation': partial(self._conv, conv_fun=int),
            'activities/minutesSedentary': partial(self._conv, conv_fun=int),
            'activities/minutesLightlyActive': partial(self._conv, conv_fun=int),
            'activities/minutesFairlyActive': partial(self._conv, conv_fun=int),
            'activities/minutesVeryActive': partial(self._conv, conv_fun=int),
            'activities/activityCalories': partial(self._conv, conv_fun=int),
            # Body
            'body/bmi': partial(self._conv, conv_fun=float),
            'body/fat': partial(self._conv, conv_fun=float),
            'body/weight': partial(self._conv, conv_fun=float),
            # Foods
            'foods/log/caloriesIn': partial(self._conv, conv_fun=int),
            'foods/log/water': partial(self._conv, conv_fun=float),
            # Sleep
            'sleep/awakeningsCount': partial(self._conv, conv_fun=int),
            'sleep/efficiency': partial(self._conv, conv_fun=int),
            'sleep/minutesAfterWakeup': partial(self._conv, conv_fun=int),
            'sleep/minutesAsleep': partial(self._conv, conv_fun=int),
            'sleep/minutesAwake': partial(self._conv, conv_fun=int),
            'sleep/minutesToFallAsleep': partial(self._conv, conv_fun=int),
            'sleep/startTime': partial(self._conv, conv_fun=None),
            'sleep/timeInBed': partial(self._conv, conv_fun=int)
        }

    def _conv(self, resource, conv_fun):
        resp = self._api_fetch_current(resource)
        return conv_fun(resp) if conv_fun else resp

    def _api_fetch_current(self, resource):
        resp = self.client.time_series(resource, period='7d')
        first_key = next(iter(resp.values()))
        if first_key is None:
            return None  # pragma: no cover
        return first_key[-1].get('value')

    def _call(self, resource):
        return self._resource_map.get(resource)(resource)

    def poll(self):
        p = Parallel(workers=4)
        for r in self._resources:
            p(self._call, fun_key=r, resource=r)
        p.run_until_complete()
        return transform_dict_items(p.results, keys_fun=camel_to_snake)



from . import UserDefinedFunction


class State (UserDefinedFunction):
    __prefix__ = 'hass'

    def __init__(self, url, token, timeout=10, **kwargs):
        super().__init__(**kwargs)
        self.url = str(url)
        self.token = str(token)
        self.timeout = timeout and int(timeout)
        if self.timeout <= 0:
            self.timeout = None  # Basically means no timeout

    def action(self, entity_id, attribute=None):
        from requests import get
        import urllib.parse as urlparse

        entity_id = str(entity_id)
        attribute = attribute and str(attribute)
        url = urlparse.urljoin(self.url, '/api/states/{entity_id}'.format(**locals()))
        headers = {
            'Authorization': 'Bearer {self.token}'.format(**locals()),
            'content-type': 'application/json',
        }

        response = get(url, headers=headers, timeout=self.timeout)
        if response.status_code != 200:
            raise RuntimeError("Failed to fetch the state for {entity_id} @ {self.url}"
                               "\nHttp Code: {response.status_code}"
                               "\nMessage: {response.text}".format(**locals()))

        if attribute:
            return response.json().get('attributes', {}).get(str(attribute))

        return response.json().get('state', None)



import os


class Validator:
    @staticmethod
    def _allow_none(arg_value, allow_none):
        return not (allow_none and arg_value is None)

    @staticmethod
    def cast_or_none(cast_fun, arg_value):
        """
        Examples:
            >>> print(Validator.cast_or_none(str, None))
            None
            >>> Validator.cast_or_none(str, 1)
            '1'
            >>> Validator.cast_or_none(int, 'a')
            Traceback (most recent call last):
            ...
            ValueError: invalid literal for int() with base 10: 'a'
        """
        return cast_fun(arg_value) if arg_value is not None else None

    @staticmethod
    def is_not_none(**kwargs):
        """
        Examples:

            >>> Validator.is_not_none(arg1="foo")  # Ok
            >>> Validator.is_not_none(arg1=None)  # None
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg1' is expected to be not none
        """
        for arg_name, arg_value in kwargs.items():
            if arg_value is None:
                raise ValueError("Argument '{arg_name}' is expected to be not none".format(**locals()))

    @staticmethod
    def one_not_none(**kwargs):
        """
        Examples:
            >>> Validator.one_not_none(arg1=None, arg2=None, arg3='passed')
            >>> Validator.one_not_none(arg1=None, arg2=None, arg3=None)
            Traceback (most recent call last):
            ...
            ValueError: Arguments ['arg1', 'arg2', 'arg3'] expects at least one passed, but all are none
        """
        if all([x is None for x in kwargs.values()]):
            raise ValueError("Arguments {args} expects at least one passed, but all are none"
                             .format(args=sorted(list(kwargs.keys()))))

    @staticmethod
    def is_instance(*required_type, allow_none=False, **kwargs):
        """
        Examples:

            >>> Validator.is_instance(str, arg="i am a string") #  Should be ok
            >>> Validator.is_instance(str, allow_none=True, arg=None)  # Should be ok as well
            >>> Validator.is_instance(bool, arg="i am a string")  # Not ok
            Traceback (most recent call last):
            ...
            TypeError: Argument 'arg' is expected to be a (<class 'bool'>,), but is <class 'str'>
            >>> Validator.is_instance(bool, str, arg="i am a string")  # Should be ok, too
            >>> Validator.is_instance(str, bool, arg1="i am a string", arg2=True)  # Ok ...
            >>> Validator.is_instance(str, arg1="i am a string", arg2=True)  # Not ok
            Traceback (most recent call last):
            ...
            TypeError: Argument 'arg2' is expected to be a (<class 'str'>,), but is <class 'bool'>
        """
        for arg_name, arg_value in kwargs.items():
            if Validator._allow_none(arg_value, allow_none) and not isinstance(arg_value, required_type):
                raise TypeError("Argument '{arg_name}' is expected to be a {required_type}"
                                ", but is {actual_type}".format(**locals(), actual_type=type(arg_value)))

    @staticmethod
    def is_non_negative(allow_none=False, **kwargs):
        """
        Examples:
            >>> Validator.is_non_negative(arg=0)
            >>> Validator.is_non_negative(arg=1)
            >>> Validator.is_non_negative(arg=-0.01)
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg' is expected to be greater or equal to zero, but it is not
            >>> Validator.is_non_negative(arg="a")
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg' is not numeric (float, int, ...)
        """
        for arg_name, arg_value in kwargs.items():
            try:
                arg_value = float(arg_value)
            except (TypeError, ValueError):
                raise ValueError("Argument '{arg_name}' is not numeric (float, int, ...)".format(**locals()))

            if Validator._allow_none(arg_value, allow_none) and arg_value < 0.0:
                raise ValueError("Argument '{arg_name}' is expected to be greater or equal to zero"
                                 ", but it is not".format(**locals()))

    @staticmethod
    def one_of(possible, allow_none=False, **kwargs):
        """
        Examples:
            >>> Validator.one_of(['a', 'b', 'c'], arg='a')
            >>> Validator.one_of(['a', 'b', 'c'], allow_none=True, arg='a', arg2=None)  # Should be ok as well
            >>> Validator.one_of(['a', 'b', 'c'], arg='z')
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg' is expected to be one of ['a', 'b', 'c'], but is 'z'
            >>> Validator.one_of(['a', 'b', 'c'], arg1='a', arg2='z')
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg2' is expected to be one of ['a', 'b', 'c'], but is 'z'
        """
        Validator.is_instance(list, tuple, possible=possible)
        for arg_name, arg_value in kwargs.items():
            if Validator._allow_none(arg_value, allow_none) and arg_value not in possible:
                raise ValueError("Argument '{arg_name}' is expected to be one of {possible}, "
                                 "but is '{arg_value}'".format(**locals()))

    @staticmethod
    def subset_of(possible, allow_none=False, **kwargs):
        """
        Examples:

            >>> Validator.subset_of(['a', 'b', 'c'], arg='a')
            >>> Validator.subset_of(['a', 'b', 'c'], arg=('a', 'b'))
            >>> Validator.subset_of(['a', 'b', 'c'], arg=('a', 'b'), arg2='c')
            >>> Validator.subset_of(['a', 'b', 'c'], arg='d')
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg' is expected to be a subset of ['a', 'b', 'c'], but is 'd'
            >>> Validator.subset_of(['a', 'b', 'c'], allow_none=True, arg1=None, arg2=None)
        """
        from .utils import make_list
        Validator.is_instance(list, tuple, possible=possible)
        for arg_name, arg_value in kwargs.items():
            if Validator._allow_none(arg_value, allow_none) and not (set(make_list(arg_value)) <= set(possible)):
                raise ValueError("Argument '{arg_name}' is expected to be a subset of {possible}, but is '{arg_value}'"
                                 .format(**locals()))

    @staticmethod
    def is_directory(allow_none=False, **kwargs):
        """
        Examples:
            >>> Validator.is_directory(arg='/tmp')
            >>> Validator.is_directory(arg='/thisonedoesnotexists')
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg' is expected to be a directory, but is '/thisonedoesnotexists'
        """
        for arg_name, arg_value in kwargs.items():
            if Validator._allow_none(arg_value, allow_none) and not os.path.isdir(arg_value):
                raise ValueError("Argument '{arg_name}' is expected to be a directory, "
                                 "but is '{arg_value}'".format(**locals()))

    @staticmethod
    def is_file(allow_none=False, **kwargs):
        """
        Examples:
            >>> import tempfile
            >>> with tempfile.NamedTemporaryFile() as tmpf:
            ...     Validator.is_file(arg=tmpf.name)
            >>> Validator.is_file(arg='/thisonedoesnotexists.txt')
            Traceback (most recent call last):
            ...
            ValueError: Argument 'arg' is expected to be a file, but is '/thisonedoesnotexists.txt'
        """
        for arg_name, arg_value in kwargs.items():
            if Validator._allow_none(arg_value, allow_none) and not os.path.isfile(arg_value):
                raise ValueError("Argument '{arg_name}' is expected to be a file, "
                                 "but is '{arg_value}'".format(**locals()))

    @staticmethod
    def is_function(allow_none=False, **kwargs):
        """
        Examples:
            >>> def foo():
            ...     pass
            >>> Validator.is_function(foo=foo)
            >>> Validator.is_function(bar='bar')
            Traceback (most recent call last):
            ...
            TypeError: Argument 'bar' is expected to be a function/callable, but is '<class 'str'>'
            >>> Validator.is_function(baz=lambda: True)
        """
        for arg_name, arg_value in kwargs.items():
            if Validator._allow_none(arg_value, allow_none) and not callable(arg_value):
                arg_type = type(arg_value)
                raise TypeError("Argument '{arg_name}' is expected to be a function/callable, "
                                "but is '{arg_type}'".format(**locals()))

    @staticmethod
    def is_iterable_but_no_str(allow_none=False, **kwargs):
        """
        Examples:
            >>> Validator.is_iterable_but_no_str(l=[])
            >>> Validator.is_iterable_but_no_str(t=("a", "b"))
            >>> Validator.is_iterable_but_no_str(s="abc")
            Traceback (most recent call last):
            ...
            TypeError: Argument 's' is expected to be a non-str iterable, but is '<class 'str'>'

        """
        from .utils import is_iterable_but_no_str
        for arg_name, arg_value in kwargs.items():
            if Validator._allow_none(arg_value, allow_none) and not is_iterable_but_no_str(arg_value):
                arg_type = type(arg_value)
                raise TypeError("Argument '{arg_name}' is expected to be a non-str iterable, "
                                "but is '{arg_type}'".format(**locals()))



from . import PushBase
from .. import load_optional_module
from ...utils import auto_str_ignore


@auto_str_ignore(['api_key'])
class Pushbullet (PushBase):
    __prefix__ = 'pushbullet'

    EXTRA = 'pushbullet'

    def __init__(self, api_key, title='pnp', **kwargs):
        super().__init__(**kwargs)
        self.api_key = str(api_key)
        self.title = str(title)

    def _load_deps(self):
        pb = load_optional_module('pushbullet', self.EXTRA)
        try:
            import urlparse
        except:  # For Python 3
            import urllib.parse as urlparse

        return pb, urlparse

    def _guess_mimetype(self, url):
        import mimetypes
        _, up = self._load_deps()
        res = up.urlparse(url)
        if not res.path:
            return None
        mt, _ = mimetypes.guess_type(res.path)
        return mt

    def _safe_basename(self, name):
        import os
        try:
            return os.path.basename(name)
        except:
            return name

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)

        pb, up = self._load_deps()
        client = pb.Pushbullet(self.api_key)
        pr = up.urlparse(real_payload)
        self.logger.info("[{self.name}] Sending message '{real_payload}' to Pushbullet".format(**locals()))
        if pr.scheme and pr.netloc:
            # Is URL
            self.logger.debug("[{self.name}] Payload '{real_payload}' is an url".format(**locals()))
            mt = self._guess_mimetype(real_payload)
            if not mt:
                # Some link
                self.logger.debug("[{self.name}] Payload '{real_payload}' has no mimetype associated"
                                  .format(**locals()))
                client.push_link(self.title, real_payload)
            else:
                # Some file
                self.logger.debug("[{self.name}] Guessed type '{mt}' for payload '{real_payload}'".format(**locals()))
                client.push_file(
                    file_name=self._safe_basename(pr.path),
                    file_type=mt,
                    file_url=real_payload,
                    title=self.title
                )
        else:
            # Is NOT url
            self.logger.debug("[{self.name}] Payload '{real_payload}' is a simple note".format(**locals()))
            client.push_note(self.title, real_payload)

        return payload
